// selectionmode.cpp
// by Glenn G. Chappell
// October 2003
//
// For CS 381
// OpenGL Selection Mode Demo
//
// The interesting code is mostly in doselect.
// To use this, create two new functions:
//    draw_with_names: the major part of the display function,
//                     with glLoadName calls added.
//    set_up_cords:    the gluOrtho2D (or whatever) call from
//                     the reshape function.
// To do a hit test, call doselect with the mouse position, and
//  two variables to return the hit information in (see doselect).

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff, includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 400; // Starting window width & height, in pixels
int winw, winh;               // Current window width, height. Saved by reshape.

// Keyboard
const int ESCKEY = 27;        // ASCII value of escape character


// printbitmap
// Prints the given string at the given raster position
//  using GLUT bitmap fonts.
// You probably don't want any rotations in the model/view
//  transformation when calling this function.
void printbitmap(const string msg, double x, double y)
{
   glRasterPos2d(x, y);
   for (string::const_iterator ii = msg.begin();
        ii != msg.end();
        ++ii)
   {
      glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *ii);
   }
}


// draw_with_names
// Draws the scene.
// Loads the proper name (with glLoadName - name stack must not be empty!)
//  before each object.
// Called by display (for display) and doselect (in selection mode).
void draw_with_names()
{
   glLoadName(1);
   glColor3d(0.9, 0.9, 0.1);
   glBegin(GL_TRIANGLES);
      glVertex2d(-0.8, 0.5);
      glVertex2d(0.2, 0.9);
      glVertex2d(0.2, 0.1);
   glEnd();

   glLoadName(2);
   glColor3d(0.1, 0.9, 0.5);
   glBegin(GL_TRIANGLES);
      glVertex2d(-0.2, 0.5);
      glVertex2d(0.8, 0.9);
      glVertex2d(0.8, 0.1);
   glEnd();

   glLoadName(10);
   glColor3d(0.9, 0.5, 0.1);
   glBegin(GL_TRIANGLES);
      glVertex2d(-0.8, -0.5);
      glVertex2d(0.2, -0.1);
      glVertex2d(0.2, -0.9);
   glEnd();
   glColor3d(0.5, 0.1, 0.9);
   glBegin(GL_TRIANGLES);
      glVertex2d(-0.2, -0.5);
      glVertex2d(0.8, -0.1);
      glVertex2d(0.8, -0.9);
   glEnd();
}


// set_up_coords
// Does our gluOrtho2D to set up a coordinate system.
// Assumes winw, winh have been set up by reshape.
// Does NOT set the matrix mode or load an identity.
// Called by reshape (for reshape) and doselect (in selection mode).
void set_up_coords()
{
   // We set up coordinate system so that aspect ratios are always correct,
   //  and the region from -1..1 in x & y always just fits in the viewport.
   if (winw > winh)
   {
      gluOrtho2D(-double(winw)/winh, double(winw)/winh, -1., 1.);
   }
   else
   {
      gluOrtho2D( -1., 1., -double(winh)/winw, double(winh)/winw);
   }
}


// doselect
// Use selection mode to check hits on a single pixel.
// Given:
//    x, y: pixel position in GLUT mouse-position format.
//    hitarray: array of 100 int's, alloc'd by caller (will be overwritten!).
// Returns:
//    hitcount: # of hits at (x, y) in selection mode.
//    hitarray: contains names of hits.
// Calls set_up_coords to do projection (gluOrtho2D).
// Calls draw_with_names to draw & load names.
void doselect(int x, int y, int & hitcount, int hitarray[100])
{
   const int buffersize = 512;
   GLuint selectionbuffer[buffersize];

   // Initialize selection mode
   glSelectBuffer(buffersize, selectionbuffer);
   glRenderMode(GL_SELECT);
   glInitNames();
   glPushName(0);  // Can't have an empty name stack

   // Set up the pick matrix
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();  // Save the projection
   glLoadIdentity();
   // Insert the following three lines just before
   //  your standard gluOrtho2D call.
   GLint vp[4];
   glGetIntegerv(GL_VIEWPORT, vp);
   gluPickMatrix(x, vp[3]-y, 1, 1, vp);
   set_up_coords();  // Do same gluOrtho2D call as in reshape

   // "Draw" the scene (in selection mode, so no actual output;
   //                   I'm not breaking THE RULES!)
   glMatrixMode(GL_MODELVIEW);  // Always go back to model/view mode
   draw_with_names();
   glFlush();                   // Empty the pipeline

   // Done with pick matrix
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();               // Restore the old projection
   glMatrixMode(GL_MODELVIEW);  // Always go back to model/view mode

   // Done with selection mode
   hitcount = glRenderMode(GL_RENDER);

   // Put hits into the caller's array, in a simpler format.
   GLuint * p = selectionbuffer;  // Moving ptr into selectionbuffer
   if (hitcount > 100)
      hitcount = 100;             // Can't hold more than 100 hits
   for (int whichhit=0; whichhit<hitcount; ++whichhit)
   {
      int stackdepth = *p;              // Get name stack depth for this hit
      p += (3+stackdepth);              // Skip current cell,
                                        //  min & max depth values,
                                        //  and saved stack

      if (p >= selectionbuffer + buffersize)
      {                        // Buffer overflow!
         hitcount = whichhit;  // Cut hitcount down to what we can hold 
         break;
      }

      if (stackdepth > 0)
      {
         int hitname = *(p-1);          // Name of object that was hit
         hitarray[whichhit] = hitname;  // Save name in caller's array
      }
   }
}


// display
// The GLUT display function
void display()
{
   glClear(GL_COLOR_BUFFER_BIT);

   // Draw scene
   draw_with_names();

   // Draw text instructions
   glColor3d(0., 0., 0.);  // Black text
   printbitmap("OpenGL Selection Mode Demo", -0.9, 0.9);
   printbitmap("Press left mouse button.", -0.9, 0.8);
   printbitmap("Hits are indicated in the text window.", -0.9, 0.7);
   printbitmap("Esc   Quit", -0.9, 0.6);

   glutSwapBuffers();
}


// idle
// The GLUT idle function
void idle()
{
   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
   glViewport(0, 0, w, h);
   winw = w;
   winh = h;

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   set_up_coords();

   glMatrixMode(GL_MODELVIEW);  // Always go back to model/view mode
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   }
}


// motion
// The GLUT motion function
void motion(int x, int y)
{
   int hitcount;
   int hitarray[100];

   doselect(x, y, hitcount, hitarray);

   if (hitcount == 0)
   {
      cout << "NO hits" << endl;
   }
   else  // Got a hit!
   {
      cout << "Hits:";
      for (int i=0; i<hitcount; ++i)
         cout << " " << hitarray[i];
      cout << endl;
   }
}


// mouse
// The GLUT mouse function
void mouse(int button, int state, int x, int y)
{
   motion(x, y);
}


// init
// Initializes GL states
// Called by main after window creation
void init()
{
   glClearColor(1., 1., 1., 0.);
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 381 - OpenGL Selection Mode Demo");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display); 
   glutIdleFunc(idle);
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);
   glutMouseFunc(mouse);
   glutMotionFunc(motion);

   // Do something
   glutMainLoop();

   return 0;
}
